@@ -4,7 +4,7 @@ from .layers import (
     PositionalEncoding,
     MultiheadAttention
 )
-from typing import Tuple, Optional, Dict
+from typing import Tuple, Optional, Dict, Callable
 
 
 class EncoderLayer(nn.Module):
@@ -46,8 +46,8 @@ class EncoderLayer(nn.Module):
             d_model: int,
             num_heads: int,
             dropout_rate: float,
-            ff_module: nn.Module,
-            norm_module: nn.Module
+            ff_factory: Callable[[], nn.Module],
+            norm_factory: Callable[[], nn.Module]
     ) -> None:
         super(EncoderLayer, self).__init__()
 
@@ -55,10 +55,10 @@ class EncoderLayer(nn.Module):
         self.num_heads = num_heads
         self.mha = MultiheadAttention(d_model, num_heads, dropout_rate)
 
-        self.ff_module = ff_module
+        self.ff_module = ff_factory()
 
-        self.norm1 = norm_module
-        self.norm2 = norm_module
+        self.norm1 = norm_factory()
+        self.norm2 = norm_factory()
         self.dropout = nn.Dropout(dropout_rate)
 
     def forward(self,
@@ -80,10 +80,12 @@ class EncoderLayer(nn.Module):
         :return: Encoded output tensor resulting from applying the Transformer
             encoder layer to the input tensor.
         """
+
+        nx = self.norm1(x)
         attention_output, attention_weights = self.mha(
-            x, x, x, mask, src_key_padding_mask)
-        out1 = self.norm1(x + self.dropout(attention_output))
-        out2 = self.norm2(out1 + self.ff_module(out1))
+            nx, nx, nx, mask, src_key_padding_mask)
+        out1 = x + self.dropout(attention_output)
+        out2 = out1 + self.ff_module(self.norm2(out1))
 
         return out2
 
@@ -124,21 +126,21 @@ class DecoderLayer(nn.Module):
             d_model: int,
             num_heads: int,
             dropout_rate: float,
-            ff_module: nn.Module,
-            norm_module: nn.Module
+            ff_factory: Callable[[], nn.Module],
+            norm_factory: Callable[[], nn.Module]
     ):
         super(DecoderLayer, self).__init__()
 
         self.self_attention = MultiheadAttention(d_model, num_heads, dropout_rate)
         self.cross_attention = MultiheadAttention(d_model, num_heads, dropout_rate)
 
-        self.ff_module = ff_module
+        self.ff_module = ff_factory()
 
         self.dropout1 = nn.Dropout(dropout_rate)
         self.dropout2 = nn.Dropout(dropout_rate)
-        self.norm1 = norm_module
-        self.norm2 = norm_module
-        self.norm3 = norm_module
+        self.norm1 = norm_factory()
+        self.norm2 = norm_factory()
+        self.norm3 = norm_factory()
 
     def forward(
             self,
@@ -183,19 +185,22 @@ class DecoderLayer(nn.Module):
             - cross_attention_weights: The attention weights produced by the cross-attention
               mechanism, with shape `(batch_size, num_heads, target_seq_len, source_seq_len)`.
         """
+
+        nx = self.norm1(x)
         self_attention_output, self_attention_weights = self.self_attention(
-            x, x, x, mask=tgt_mask, key_padding_mask=tgt_key_padding_mask)
-        out1 = self.norm1(x + self.dropout1(self_attention_output))
+            nx, nx, nx,
+            mask=tgt_mask, key_padding_mask=tgt_key_padding_mask)
+        out1 = x + self.dropout1(self_attention_output)
 
         if memory is not None:
             cross_attention_output, cross_attention_weights = self.cross_attention(
-                out1, memory, memory, mask=memory_mask, key_padding_mask=memory_key_padding_mask)
-            out2 = self.norm2(out1 + self.dropout2(cross_attention_output))
+                self.norm2(out1), memory, memory, mask=memory_mask, key_padding_mask=memory_key_padding_mask)
+            out2 = out1 + self.dropout2(cross_attention_output)
         else:
             cross_attention_weights = None
             out2 = out1
 
-        out3 = self.norm3(out2 + self.ff_module(out2))
+        out3 = out2 + self.ff_module(self.norm3(out2))
 
         return out3, self_attention_weights, cross_attention_weights
 
@@ -231,8 +236,9 @@ class Encoder(nn.Module):
             num_heads: int,
             input_vocab_size: int,
             dropout_rate: float,
-            ff_module: nn.Module,
-            norm_module: nn.Module
+            ff_factory: Callable[[], nn.Module],
+            norm_factory: Callable[[], nn.Module],
+            pad_token_id: int
     ) -> None:
         """
         Initializes the Encoder with specified parameters and creates required
@@ -248,14 +254,15 @@ class Encoder(nn.Module):
         """
         super(Encoder, self).__init__()
         self.pe = PositionalEncoding(d_model)
-        self.embedding = nn.Embedding(input_vocab_size, d_model, padding_idx=0)
+        self.embedding = nn.Embedding(input_vocab_size, d_model, padding_idx=pad_token_id)
         self.dropout = nn.Dropout(dropout_rate)
         self.d_model = d_model
 
         self.encoder_layers = nn.ModuleList(
-            [EncoderLayer(d_model, num_heads, dropout_rate, ff_module, norm_module)
+            [EncoderLayer(d_model, num_heads, dropout_rate, ff_factory, norm_factory)
              for _ in range(num_encoder_layers)]
         )
+        self.final_layer_norm = norm_factory()
 
     def forward(
             self,
@@ -281,11 +288,12 @@ class Encoder(nn.Module):
         :return: Processed tensor after passing through all encoder layers.
         :rtype: torch.Tensor
         """
-        x = self.embedding(x.to(torch.long)) * self.d_model ** 0.5
+        x = self.embedding(x.to(torch.long)) * (self.d_model ** 0.5)
         x = self.dropout(self.pe(x))
 
         for layer in self.encoder_layers:
             x = layer(x, src_mask, src_key_padding_mask)
+        x = self.final_layer_norm(x)
 
         return x
 
@@ -320,8 +328,9 @@ class Decoder(nn.Module):
             num_heads: int,
             target_vocab_size: int,
             dropout_rate: float,
-            ff_module: nn.Module,
-            norm_module: nn.Module
+            ff_factory: Callable[[], nn.Module],
+            norm_factory: Callable[[], nn.Module],
+            pad_token_id: int
     ) -> None:
         """
         A Decoder class that implements a multi-layer transformer decoder for sequence-to-sequence
@@ -346,20 +355,21 @@ class Decoder(nn.Module):
         :type dropout_rate: float
         """
         super(Decoder, self).__init__()
-        self.embedding = nn.Embedding(target_vocab_size, d_model, padding_idx=0)
+        self.embedding = nn.Embedding(target_vocab_size, d_model, padding_idx=pad_token_id)
         self.pe = PositionalEncoding(d_model)
         self.dropout = nn.Dropout(dropout_rate)
         self.d_model = d_model
 
         self.decoder_layers = nn.ModuleList(
-            [DecoderLayer(d_model, num_heads, dropout_rate, ff_module, norm_module)
+            [DecoderLayer(d_model, num_heads, dropout_rate, ff_factory, norm_factory)
              for _ in range(num_decoder_layers)]
         )
+        self.final_layer_norm = norm_factory()
 
     def forward(
             self,
             tgt: torch.Tensor,
-            memory: torch.Tensor,
+            memory: Optional[torch.Tensor] = None,
             tgt_mask: Optional[torch.Tensor] = None,
             memory_mask: Optional[torch.Tensor] = None,
             tgt_key_padding_mask: Optional[torch.Tensor] = None,
@@ -383,7 +393,7 @@ class Decoder(nn.Module):
         """
         attention_weights = {}
 
-        x = self.embedding(tgt.to(torch.long)) * self.d_model ** 0.5
+        x = self.embedding(tgt.to(torch.long)) * (self.d_model ** 0.5)
         x = self.dropout(self.pe(x))
 
         for i, layer in enumerate(self.decoder_layers):
@@ -401,6 +411,8 @@ class Decoder(nn.Module):
                 "decoder_layer_{}_cross_attention_weights".format(i + 1)
             ] = cross_attention_weights
 
+        x = self.final_layer_norm(x)
+
         return x, attention_weights
 
 
@@ -429,8 +441,9 @@ class Transformer(nn.Module):
             input_vocab_size: int,
             target_vocab_size: int,
             dropout_rate: float,
-            ff_module: nn.Module,
-            norm_module: nn.Module,
+            ff_factory: Callable[[], nn.Module],
+            norm_factory: Callable[[], nn.Module],
+            pad_token_id: int,
             use_encoder: bool = True
     ) -> None:
         """
@@ -449,21 +462,20 @@ class Transformer(nn.Module):
         """
         super(Transformer, self).__init__()
 
-        assert ff_module is not None, ("ff_module must be provided. Please import FeedForwardNetwork or other layers "
-                                       "from ukraine.transformers.layers.")
-        assert norm_module is not None, ("norm_module must be provided. Please import DyT from  "
-                                         "ukraine.transformers.layers or use torch.nn.LayerNorm.")
-
         self.use_encoder = use_encoder
 
         if use_encoder:
             self.encoder = Encoder(num_encoder_layers, d_model, num_heads,
-                                   input_vocab_size, dropout_rate, ff_module, norm_module)
+                                   input_vocab_size, dropout_rate,
+                                   ff_factory, norm_factory,
+                                   pad_token_id)
         else:
             self.encoder = None
         self.decoder = Decoder(num_decoder_layers, d_model, num_heads,
-                               target_vocab_size, dropout_rate, ff_module, norm_module)
-        self.output_fc = nn.Linear(d_model, target_vocab_size)
+                               target_vocab_size, dropout_rate,
+                               ff_factory, norm_factory,
+                               pad_token_id)
+        self.output_fc = nn.Linear(d_model, target_vocab_size, bias=False)
 
     def forward(
             self,
